/**
 * $Id: pluginapi.c,v 1.9 2006/09/06 19:13:23 blendix Exp $
 *
 * ***** BEGIN GPL/BL DUAL LICENSE BLOCK *****
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. The Blender
 * Foundation also sells licenses for use in proprietary software under
 * the Blender License.  See http://www.blender.org/BL/ for information
 * about this.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * The Original Code is Copyright (C) 2001-2002 by NaN Holding BV.
 * All rights reserved.
 *
 * The Original Code is: all of this file.
 *
 * Contributor(s): Peter Schlaile
 *
 * ***** END GPL/BL DUAL LICENSE BLOCK *****
 * Wrappers for the plugin api. This api is up for removal.
 */

/* There are four headers making up the plugin api:
 * - floatpatch.h : Wraps math functions for mips platforms, no code
 *                  required.
 * - iff.h        : Defines, macros and functions for dealing
 *                  with image buffer things.
 * - plugin.h     : Wraps some plugin handling types, accesses noise
 *                  functions.
 * - util.h       : Useful defines, memory management.
 */

#define PLUGIN_INTERN /* This tells the LIBEXPORT macro to compile with
	dll export set on windows */

#include "blenpluginapi/util.h"
#include "plugin.h"
#include "PLU_private.h"

#include "MEM_guardedalloc.h"

#include "BLI_blenlib.h"  /* util and noise functions */
#include "BLI_threads.h"  /* For threadsfe guardedalloc malloc/calloc/free */
#include "IMB_imbuf.h"    /* image buffer stuff       */
#include "PIL_dynlib.h"
#include "BKE_idprop.h"
#include "BKE_ipo.h"
#include "DNA_curve_types.h"

#undef CLAMP
#undef ABS

#include "BKE_utildefines.h"

#include <stdarg.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
 
/* -------------------------------------------------------------------------- */
/* stuff from util.h                                                          */ 
/* -------------------------------------------------------------------------- */

LIBEXPORT void *mallocN(int len, char *str)
{
	return MEM_mallocN(len, str);
}

LIBEXPORT void *callocN(int len, char *str)
{
	return MEM_callocN(len, str);
}

LIBEXPORT short freeN(void *vmemh)
{
	return MEM_freeN(vmemh);
}

/* these are not needed anymore, mallocN/callocN/freeN is now threadsafe */
LIBEXPORT void *mallocT(int len, char *str)
{
	return MEM_mallocN(len, str);
}

LIBEXPORT void *callocT(int len, char *str)
{
	return MEM_callocN(len, str);
}

LIBEXPORT void freeT(void *vmemh)
{
	MEM_freeN(vmemh);
	return;
}


/* -------------------------------------------------------------------------- */
/* stuff from iff.h                                                           */ 
/* -------------------------------------------------------------------------- */

LIBEXPORT struct ImBuf *allocImBuf(short x,
						 short y,
						 uchar d,
						 uint flags,
						 uchar bitmap)
{
	return IMB_allocImBuf(x, y, d, flags, bitmap);
}


LIBEXPORT struct ImBuf *dupImBuf(struct ImBuf *ib)
{
	return IMB_dupImBuf(ib);
}
	
LIBEXPORT void freeImBuf(struct ImBuf* ib)
{
	IMB_freeImBuf(ib);
}

LIBEXPORT short converttocmap(struct ImBuf* ibuf)
{
	return IMB_converttocmap(ibuf);
}

LIBEXPORT short saveiff(struct ImBuf *ib,
			  char *c,
			  int i)
{
	return IMB_saveiff(ib, c, i);
}

LIBEXPORT struct ImBuf *loadiffmem(int *mem,int flags)
{
	return IMB_loadiffmem(mem, flags);
}
	
LIBEXPORT struct ImBuf *loadifffile(int a,
						  int b)
{
	return IMB_loadifffile(a, b);
}

LIBEXPORT struct ImBuf *loadiffname(char *n,
						  int flags)
{
	return IMB_loadiffname(n, flags);
}
	
LIBEXPORT struct ImBuf *testiffname(char *n,
						  int flags)
{
	return IMB_testiffname(n, flags);
}

LIBEXPORT struct ImBuf *onehalf(struct ImBuf *ib)
{
	return IMB_onehalf(ib);
}

LIBEXPORT struct ImBuf *onethird(struct ImBuf *ib)
{
	return IMB_onethird(ib);
}

LIBEXPORT struct ImBuf *halflace(struct ImBuf *ib)
{
	return IMB_halflace(ib);
}

LIBEXPORT struct ImBuf *half_x(struct ImBuf *ib)
{
	return IMB_half_x(ib);
}

LIBEXPORT struct ImBuf *half_y(struct ImBuf *ib)
{
	return IMB_half_y(ib);
}

LIBEXPORT struct ImBuf *double_x(struct ImBuf *ib)
{
	return IMB_double_x(ib);
}

LIBEXPORT struct ImBuf *double_y(struct ImBuf *ib)
{
	return IMB_double_y(ib);
}

LIBEXPORT struct ImBuf *double_fast_x(struct ImBuf *ib)
{
	return IMB_double_fast_x(ib);
}

LIBEXPORT struct ImBuf *double_fast_y(struct ImBuf *ib)
{
	return IMB_double_fast_y(ib);
}

LIBEXPORT int ispic(char * name)
{
	return IMB_ispic(name);
}

LIBEXPORT void dit2(struct ImBuf *ib,
		   short a,
		   short b)
{
	IMB_dit2(ib, a, b);
}

LIBEXPORT void dit0(struct ImBuf *ib,
		   short a,
		   short b)
{
	IMB_dit0(ib, a, b);
}

/* still the same name */
/*  void (*ditherfunc)(struct ImBuf *, short, short){} */

LIBEXPORT struct ImBuf *scaleImBuf(struct ImBuf *ib,
						 short nx,
						 short ny)
{
	return IMB_scaleImBuf(ib, nx, ny);
}

LIBEXPORT struct ImBuf *scalefastImBuf(struct ImBuf *ib,
							 short x,
							 short y)
{
	return IMB_scalefastImBuf(ib, x, y);
}


LIBEXPORT struct ImBuf *scalefieldImBuf(struct ImBuf *ib,
							  short x,
							  short y)
{
	return IMB_scalefieldImBuf(ib, x, y);
}

LIBEXPORT struct ImBuf *scalefastfieldImBuf(struct ImBuf *ib,
								  short x,
								  short y)
{
	return IMB_scalefastfieldImBuf(ib, x, y);
}

	/* Extra ones that some NaN (read Ton) plugins use,
	 * even though they aren't in the header
	 */

LIBEXPORT void interlace(struct ImBuf *ibuf)
{
	IMB_interlace(ibuf);
}

LIBEXPORT void gamwarp(struct ImBuf *ibuf, double gamma)
{
	IMB_gamwarp(ibuf,gamma);
}
	 
LIBEXPORT void de_interlace(struct ImBuf *ib)
{
	IMB_de_interlace(ib);
}

/* -------------------------------------------------------------------------- */
/* stuff from plugin.h                                                        */ 
/* -------------------------------------------------------------------------- */
static ListBase * desc_lists = NULL;
static int plugin_first_run = 1;
static int plugin_scan_mode = 0;
static const char * current_plugin_filename = 0;
static time_t current_lastchanged = 0;

static plugin_descriptor_t * new_descriptor(plugin_type_t type,
					    unsigned long capabilities,
					    char* name,
					    char* plugin_filename,
					    char* menu_entry,
					    char* author,
					    char* copyright,
					    char* help,
					    char* date,
					    char* file_pattern,
					    time_t lastchanged)
{
	plugin_descriptor_t * d = (plugin_descriptor_t *)
		callocN(sizeof(plugin_descriptor_t), "plugin_descriptor");
	d->plugin_type = type;
	d->capabilities = capabilities;
	d->name = name;
	d->plugin_filename = plugin_filename;
	d->menu_entry = menu_entry;
	d->author = author;
	d->copyright = copyright;
	d->help = help;
	d->date = date;
	d->file_pattern = file_pattern;
	d->lastchanged = lastchanged;

	return d;
}

static void delete_descriptor(plugin_descriptor_t * d)
{
	if (d->name) { freeN(d->name); }
	if (d->plugin_filename) { freeN(d->plugin_filename); }
	if (d->menu_entry) { freeN(d->menu_entry); }
	if (d->author) { freeN(d->author); }
	if (d->copyright) { freeN(d->copyright); }
	if (d->help) { freeN(d->help); }
	if (d->date) { freeN(d->date); }
	if (d->file_pattern) { freeN(d->file_pattern); }

	freeN(d);
}

static void load_plugin(const char* filename)
{
	PILdynlib* h = PIL_dynlib_open((char*) filename);
	void (*query_func)(); 
	struct stat status;

	if (!h) {
		return;
	}

	stat(filename, &status);

	current_plugin_filename = filename;
	current_lastchanged = status.st_mtime;

	query_func = (void (*)()) PIL_dynlib_find_symbol(h, "plugin_query");

	if (query_func) {
		query_func();

		PIL_dynlib_close(h);

		return;
	} 

	/* FIXME: fallback code for old plugins */

	return;
}

static void load_plugin_descriptor(plugin_descriptor_t * d)
{
	if (d->interface_version != 0) {
		return;
	}

	load_plugin(d->plugin_filename);
}


static void init_desc_lists()
{
	if (!desc_lists) {
		desc_lists = (ListBase *) 
			callocN(sizeof(ListBase) * PLUGIN_SIZEOF_TYPE,
				"plugin_descriptor array");
	}
}

static plugin_descriptor_t * get_desc_list(plugin_type_t type)
{
	init_desc_lists();

	return (plugin_descriptor_t *) desc_lists[type].first;
}

static void insert_desc_entry(plugin_descriptor_t * t)
{
	plugin_type_t type = t->plugin_type;
	plugin_descriptor_t * it = get_desc_list(type);

	while (it) {
		if (strcmp(it->name, t->name) == 0) {
			break;
		}
		it = it->next;
	}

	if (!it) {
		BLI_addtail(&desc_lists[type], t);
	} else {
		if (t->interface_version || !it->interface_version) {
			BLI_insertlink(&desc_lists[type], it, t);
			BLI_remlink(&desc_lists[type], it);

			delete_descriptor(it);
		} else {
			delete_descriptor(t);
		}
	}
}


/* ---------------------------------------------------------------------- */
/* - plugin cache file / plugin scan code - */

#define PLUGIN_CACHE_FILE_VERSION 1
#define PLUGIN_CACHE_STRING_MAX   1000
#define CACHE_DATAFILE "Bplugins"
#define MAX_DIR_DEPTH   4 /* max depth for traversing plugin dirs */
#define MAX_DIR_NUMBER 30 /* max number of dirs in plugin dirs trees */

static void write_keyval_str(FILE * fp, const char * key, const char * str)
{
	fprintf(fp, "%s ", key);

	while (*str) {
		switch (*str) {
		case '\\':
			fputc('\\', fp);
			fputc('\\', fp);
			break;
		case '\n':
			fputc('\\', fp);
			fputc('n', fp);
			break;
		default:
			fputc(*str, fp);
			break;
		}
	}
	fputc('\n', fp);
}

static void write_keyval_int(FILE * fp, const char * key, 
			     unsigned long val)
{
	fprintf(fp, "%s %ld\n", key, val);
}

static unsigned long read_keyval_int(FILE *fp, char * key)
{
	unsigned long rv;
	int a;
	int i;

	for (i = 0; i < PLUGIN_CACHE_STRING_MAX; i++) {
		a = fgetc(stdin);
		
		if (a == ' ') {
			break;
		}
		if (a == EOF || a == '\n') {
			return 0;
		}
		*key++ = a;
	}

	fscanf(fp, "%ld", &rv);

	for (;;) {
		int a = fgetc(fp);

		if (a == '\n') {
			break;
		}
		if (a == EOF) {
			break;
		}
	}

	return rv;
}

static char * read_keyval_str(FILE * fp, char * key)
{
	char buf[PLUGIN_CACHE_STRING_MAX];
	char * p = buf;
	char * rv;
	int len;

	fgets(buf, PLUGIN_CACHE_STRING_MAX, fp);

	while (*p && *p != ' ') {
		*key++ = *p;
		p++;
	}

	if (!*p) {
		return 0;
	}

	p++;

	len = strlen(p);

	rv = (char*) mallocN(len + 1, "plugin descriptor string");

	memcpy(rv, p, len + 1);

	if (p[len-1] == '\n') {
		rv[len - 1] = 0;
		return rv;
	}

	for (;;) {
		int a = fgetc(fp);

		if (a == '\n') {
			break;
		}
		if (a == EOF) {
			break;
		}
	}

	return rv;
}

static void serialize_desc_lists(const char * filename)
{
	FILE * fp = fopen(filename, "w");
	int i;

	if (!fp) {
		return;
	}

	init_desc_lists();

	write_keyval_int(fp, "Blender-Plugin-Cache", 
			 PLUGIN_CACHE_FILE_VERSION);

	fputc('\n', fp);

	for (i = 0; i < PLUGIN_SIZEOF_TYPE; i++) {
		plugin_descriptor_t * d = get_desc_list(i);

		while (d) {
			write_keyval_int(fp, "type", d->plugin_type);
			write_keyval_str(fp, "name", d->name);
			write_keyval_str(fp, "plugin_filename", 
					 d->plugin_filename);
			write_keyval_str(fp, "menu_entry", 
					 d->menu_entry);
			write_keyval_str(fp, "author", 
					 d->author);
			write_keyval_str(fp, "copyright", 
					 d->copyright);
			write_keyval_str(fp, "help", 
					 d->date);
			write_keyval_str(fp, "file_pattern", 
					 d->file_pattern);
			write_keyval_int(fp, "lastchanged", d->lastchanged);

			fputc('\n', fp);

			d = d->next;
		}
	}

	fclose(fp);
}

static int unserialize_desc_lists(const char * filename)
{
	FILE * fp = fopen(filename, "w");
	int i;

	if (!fp) {
		return 0;
	}

	/* FIXME: write unserialize code */
	
	fclose(fp);

	return 0;
}

static void plugin_scan_file(char* filename)
{
	plugin_scan_mode = 1;

	load_plugin(filename);

	plugin_scan_mode = 0;
}

static int Dir_Depth;
static int Dirs_Number;

static int plugin_parsedir(const char* dirname, const char* parentdir)
{
	DIR *dir; 
	struct dirent *de;
	struct stat status;
	char path[FILE_MAX];
	char subdir[FILE_MAX];
	char *s = NULL;
	
	dir = opendir(dirname);

	if (dir != NULL) {
		while ((de = readdir(dir)) != NULL) {

			/* skip files and dirs starting with '.' */
			if (de->d_name[0] == '.') {
				continue;
			}
			
			BLI_make_file_string("/", path, dirname, de->d_name);
			
			if (stat(path, &status) != 0) {
				continue;
			}

			if (S_ISREG(status.st_mode)) { /* is file */
				plugin_scan_file(path);
			} else if (S_ISDIR(status.st_mode)) { /* is subdir */
				Dirs_Number++;
				Dir_Depth++;
				if (Dirs_Number > MAX_DIR_NUMBER) {
					closedir(dir);
					return -1;
				}
				else if (Dir_Depth > MAX_DIR_DEPTH) {
					closedir(dir);
					return -1;
				}
				s = de->d_name;
				if (parentdir) {
					/* Join parentdir and de->d_name */
					BLI_join_dirfile(subdir, parentdir, 
							 de->d_name);
					s = subdir;
				}
				if (plugin_parsedir(path, s) == -1) {
					closedir(dir);
					return -1;
				}
				Dir_Depth--;
			}

		}
		closedir(dir);
	} else { 
		return -1;
	}

	return 0;
}

static void plugin_scan_all(int force_refresh)
{
	const char * cache_file = "/tmp/BlenCache";
	const char * plugin_dir = "/tmp/PluginDir";

	if (!plugin_first_run && !force_refresh) {
		return;
	}

	if (force_refresh) {
		plugin_parsedir(plugin_dir, 0);
		serialize_desc_lists(cache_file);
	} else {
		if (!unserialize_desc_lists(cache_file)) {
			plugin_parsedir(plugin_dir, 0);
			serialize_desc_lists(cache_file);
		}

		plugin_first_run = 0;
	}
}

/* ---------------------------------------------------------------------- */
/* - property handling - */

LIBEXPORT int plugin_bind_property(plugin_control_t * control,
				   int prop_type, const char* name, 
				   void * data)
{
	IDProperty* idp = IDP_GetPropertyFromGroup(control->p_data, name);

	if (!idp) {
		IDPropertyTemplate tmpl;
		int type;
	
		switch(prop_type) {
		case PLP_STRING:
			tmpl.str = (char*) data;
			type = IDP_STRING;
			break;
		case PLP_INT:
			tmpl.i = *(int*) data;
			type = IDP_INT;
			break;
		case PLP_FLOAT:
			tmpl.f = *(float*) data;
			type = IDP_FLOAT;
			break;
		default:
			return 0;
		}

		idp = IDP_New(type, tmpl, name);
		IDP_AddToGroup(control->p_data, idp);
	} else {
		switch (prop_type) {
		case PLP_INT:
			if (idp->type != IDP_INT) {
				return 0;
			}
			break;
		case PLP_FLOAT:
			if (idp->type != IDP_FLOAT) {
				return 0;
			}
			break;
		case PLP_STRING:
			if (idp->type != IDP_STRING) {
				return 0;
			}
			break;
		default:
			return 0;
		}
	}
	IDP_BindProperty(idp, data);
	return 1;
}

LIBEXPORT int plugin_bind_properties(
	plugin_control_t * control,
	int prop_type, const char * name, void * data, ...)
{
	va_list ap;

	int rv = 0;

	rv += plugin_bind_property(control, prop_type, name, data);

	va_start(ap, data);

	for (;;) {
		prop_type = va_arg(ap, int);
		if (prop_type == PLP_END) {
			break;
		}
		name = va_arg(ap, const char*);
		data = va_arg(ap, void*);

		rv += plugin_bind_property(control, prop_type, name, data);
	}

	va_end(ap);

	return rv;
}

LIBEXPORT int plugin_bind_properties_varstruct(
	plugin_control_t * control,
	VarStruct * vstruct,
	unsigned int num_elems, 
	void * begin_of_data)
{
	int rv = 0;
	unsigned int i;
	int * p = (int*) begin_of_data;

	for (i = 0; i < num_elems; i++) {
		if ((vstruct->type & INT) == INT) {
			rv += plugin_bind_property(
				control, PLP_INT, vstruct->name, p);
		} else if ((vstruct->type & FLO) == FLO) {
			rv += plugin_bind_property(
				control, PLP_FLOAT, vstruct->name, p);
		}
		p++;
		vstruct++;
	}
	return rv;
}

/* returns false if property not available! */
LIBEXPORT int plugin_get_property(plugin_control_t * control,
				  int prop_type, const char * name, 
				  void * data)
{
	IDProperty* idp = IDP_GetPropertyFromGroup(control->p_data, name);

	if (!idp) {
		return 0;
	}

	switch(prop_type) {
	case PLP_STRING:
		if (idp->type == IDP_STRING) {
			*(char**) data = IDP_String(idp);
		} else {
			return 0;
		}
		break;
	case PLP_INT:
		switch(idp->type) {
		case IDP_INT:
			*(int*) data = IDP_Int(idp);
			break;
		case IDP_FLOAT:
			*(int*) data = (int) IDP_Float(idp);
			break;
		default:
			return 0;
		}
		break;
	case PLP_FLOAT:
		switch(idp->type) {
		case IDP_INT:
			*(float*) data = (float) IDP_Int(idp);
			break;
		case IDP_FLOAT:
			*(float*) data = IDP_Float(idp);
			break;
		default:
			return 0;
		}
		break;
	default:
		return 0;
	}

	return 1;
}

LIBEXPORT int plugin_set_property(plugin_control_t * control,
				  plp_type_t prop_type, const char * name, 
				  void * data)
{
	IDPropertyTemplate tmpl;
	IDProperty* idp;

	int type;

	switch(prop_type) {
	case PLP_STRING:
		tmpl.str = (char*) data;
		type = IDP_STRING;
		break;
	case PLP_INT:
		tmpl.i = *(int*) data;
		type = IDP_INT;
		break;
	case PLP_FLOAT:
		tmpl.f = *(float*) data;
		type = IDP_FLOAT;
		break;
	default:
		return 0;
	}

	idp = IDP_GetPropertyFromGroup(control->p_data, name);
	if (!idp) {
		idp = IDP_New(type, tmpl, name);
		IDP_AddToGroup(control->p_data, idp);
	} else {
		switch(prop_type) {
		case PLP_STRING:
			if (idp->type == IDP_STRING) {
				IDP_AssignString(idp, tmpl.str);
			} else {
				return 0;
			}
			break;
		case PLP_INT:
			switch (idp->type) {
			case IDP_INT:
				IDP_Int(idp) = tmpl.i;
				break;
			case IDP_FLOAT:
				IDP_Float(idp) = (float) tmpl.i;
				break;
			default:
				return 0;
			}
		case PLP_FLOAT:
			switch (idp->type) {
			case IDP_INT:
				IDP_Int(idp) = (int) tmpl.f;
				break;
			case IDP_FLOAT:
				IDP_Float(idp) = tmpl.f;
				break;
			default:
				return 0;
			}
		default:
			return 0;
		}
	}
	return 1;
}

static int plugin_set_property_vararg(plugin_control_t * c, va_list ap)
{
	int rv = 0;
	int prop_type;
	const char * name;
	void * data;

	for (;;) {
		prop_type = va_arg(ap, int);
		if (prop_type == PLP_END) {
			break;
		}
		name = va_arg(ap, const char*);
		data = va_arg(ap, void*);

		rv += plugin_set_property(c, prop_type, name, data);
	}
	return rv;
}

/* ---------------------------------------------------------------------- */
/* - gui / ipo interface - */


LIBEXPORT void plugin_gui_draw_slider(plugin_control_t * control,
			    int type, const char* name,
			    float def, float min, float max,
			    const char* tip, 
			    float * data, int xpos, int ypos)
{
	if (!control->uplink_handler 
	    || !control->uplink_handler->gui_draw_slider) {
		return;
	}

	control->uplink_handler->gui_draw_slider(
			control, type, name,
			def, min, max,
			tip, data, xpos, ypos);
}

static struct IpoCurve * get_icu(plugin_control_t * control, 
				 const char * icuname)
{
	if (!control->uplink_handler || !control->uplink_handler->get_icu) {
		return 0;
	}

	struct IpoCurve * icu = control->uplink_handler->get_icu(
		control, icuname);

	return icu;
}

LIBEXPORT void plugin_gui_bind_ipo(plugin_control_t * control, 
				   const char * icuname,
				   float * data)
{
	struct IpoCurve * icu = get_icu(control, icuname);
	if (!icu) { return; }
	
	/* FIXME: easiest way would be extension of IPO system,
	   so that ICUs can store their values to arbitrary places...
	*/
}


LIBEXPORT void plugin_gui_icu_set_range(plugin_control_t * control, 
					const char * icuname,
					float ymin, float ymax)
{
	struct IpoCurve * icu = get_icu(control, icuname);
	if (!icu) { return; }

	icu->ymin = ymin;
	icu->ymax = ymax;
}

LIBEXPORT float plugin_gui_get_ipo_value(plugin_control_t * control, 
					 const char * icuname,
					 float frame)
{
	struct IpoCurve * icu = get_icu(control, icuname);
	float bak;
	float rv;

	if (!icu) { return 0; }

	bak = icu->curval;
	calc_icu(icu, frame_to_float(frame));
	rv = icu->curval;
	icu->curval = bak;

	return rv;
}


/* ---------------------------------------------------------------------- */
/* - core interface - */

LIBEXPORT plugin_descriptor_t * plugin_enumerate(
	plugin_type_t type, int force_refresh)
{
	plugin_scan_all(force_refresh);

	return get_desc_list(type);
}

LIBEXPORT plugin_descriptor_t * plugin_get_by_name(
	plugin_type_t type, const char* name)
{
	plugin_descriptor_t * t;

	plugin_scan_all(0);

	t = get_desc_list(type);

	while (t) {
		if (strcmp(name, t->name) == 0) {
			break;
		}
		t = t->next;
	}

	return t;
}

LIBEXPORT plugin_control_t * plugin_init_control(
	plugin_descriptor_t * d, 
	plugin_uplink_handler_t * uplink_handler,
	void * uplink,
	int prop_type, const char * name, void * data, ...)
{
	va_list ap;
	IDPropertyTemplate val;
	plugin_control_t * rv = (plugin_control_t*) callocN(
		sizeof(plugin_control_t), "plugin control");
	
	strncpy(rv->name, d->name, sizeof(rv->name)-1);
	rv->name[sizeof(rv->name)-1] = 0;

	rv->plugin_type = d->plugin_type;
	rv->descriptor = d;
	rv->uplink_handler = uplink_handler;
	rv->uplink = uplink;
	rv->p_data = IDP_New(IDP_GROUP, val, "PluginParamGroup");

	plugin_set_property(rv, prop_type, name, data);
	
	va_start(ap, data);
	plugin_set_property_vararg(rv, ap);
	va_end(ap);

	rv->instance = d->common.init(rv);

	return rv;
}

LIBEXPORT int plugin_load_control(plugin_control_t * c, 
				  plugin_uplink_handler_t * uplink_handler,
				  void * uplink)
{
	if (c->instance) {
		return 1;
	}
	c->descriptor = plugin_get_by_name(c->plugin_type, c->name);
	if (!c->descriptor) {
		return 0;
	}
	c->uplink = uplink;
	c->uplink_handler = uplink_handler;
	c->instance = c->descriptor->common.load(c);

	if (!c->instance) {
		return 0;
	}

	return 1;
}

LIBEXPORT void plugin_free_control(plugin_control_t * c)
{
	c->descriptor->common.free(c->instance);
	IDP_FreeProperty(c->p_data);
	freeN(c->p_data);
	freeN(c);
}

/* ---------------------------------------------------------------------- */
/* - plugin interface - */

LIBEXPORT plugin_descriptor_t * plugin_register_private(
	int interface_version,
	plugin_type_t type,
	unsigned long capabilities,
	const char* name,
	const char* menu_entry,
	const char* author,
	const char* copyright,
	const char* help,
	const char* date,
	const char* file_pattern,
	plugin_private_descriptor_t * desc,
	plugin_instance_t* (*init)(plugin_control_t * c),
	plugin_instance_t* (*load)(plugin_control_t * c),
	void (*free)(plugin_instance_t* This))
{
	plugin_descriptor_t * rv = new_descriptor(
		type, capabilities,
		name ? BLI_strdup((char*) name) : 0, 
		current_plugin_filename ? 
		BLI_strdup((char*) current_plugin_filename) : 0,
		menu_entry ? BLI_strdup((char*) menu_entry) : 0,
		author ? BLI_strdup((char*) author) : 0,
		copyright ? BLI_strdup((char*) copyright) : 0,
		help ? BLI_strdup((char*) help) : 0,
		date ? BLI_strdup((char*) date) : 0,
		file_pattern ? BLI_strdup((char*) file_pattern) : 0,
		current_lastchanged);

	if (!plugin_scan_mode) {
		rv->interface_version = interface_version;
		rv->desc = desc;
		rv->common.init = init;
		rv->common.free = free;
		rv->common.load = load;
	}

	insert_desc_entry(rv);

	return rv;
}





/* These three need to be defined in the plugion itself. The plugin
 * loader looks for these functions to check whether it can use the
 * plugin. For sequences, something similar exists. */
/*  int plugin_tex_getversion(void); */
/*  int plugin_seq_getversion(void); */
/*  void plugin_getinfo(PluginInfo *); */

LIBEXPORT float hnoise(float noisesize,
			 float x,
			 float y,
			 float z)
{
	return BLI_hnoise(noisesize, x, y, z);
}

LIBEXPORT float hnoisep(float noisesize,
			  float x,
			  float y,
			  float z)
{
	return BLI_hnoisep(noisesize, x, y, z);
}

LIBEXPORT float turbulence(float noisesize,
				 float x,
				 float y,
				 float z,
				 int depth)
{
	return BLI_turbulence(noisesize, x, y, z, depth);
}

LIBEXPORT float turbulence1(float noisesize,
				  float x,
				  float y,
				  float z,
				  int depth)
{
	return BLI_turbulence1(noisesize, x, y, z, depth);
}

/* -------------------------------------------------------------------------- */

	/* Stupid hack - force the inclusion of all of the
	 * above functions in the binary by 'using' each one...
	 * Otherwise they will not be imported from the archive
	 * library on Unix. -zr
	 */
int pluginapi_force_ref(void); 

int pluginapi_force_ref(void) 
{
	return (int) mallocN +
		(int) callocN +
		(int) freeN +
		(int) mallocT +
		(int) callocT +
		(int) freeT +
		(int) allocImBuf +
		(int) dupImBuf +
		(int) freeImBuf +
		(int) converttocmap +
		(int) saveiff +
		(int) loadiffmem +
		(int) loadifffile +
		(int) loadiffname +
		(int) testiffname +
		(int) onehalf +
		(int) onethird +
		(int) halflace +
		(int) half_x +
		(int) half_y +
		(int) double_x +
		(int) double_y +
		(int) double_fast_x +
		(int) double_fast_y +
		(int) ispic +
		(int) dit2 +
		(int) dit0 +
		(int) scaleImBuf +
		(int) scalefastImBuf +
		(int) scalefieldImBuf +
		(int) scalefastfieldImBuf +
		(int) hnoise +
		(int) hnoisep +
		(int) turbulence +
		(int) turbulence1 +
		(int) de_interlace +
		(int) interlace +
		(int) gamwarp;
}
